package main

import (
	"context"
	"errors"
	"flag"
	"fmt"
	"io"
	"net/http"
	"os"
	"os/signal"
	"strings"
	"sync"
	"syscall"
	"time"

	"github.com/open-io/oio-exporter/beanstalkd"
	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
	hc "github.com/open-io/oio-exporter/healthcheck"
	self "github.com/open-io/oio-exporter/internal"
	clog "github.com/open-io/oio-exporter/log"
	"github.com/open-io/oio-exporter/memcached"
	"github.com/open-io/oio-exporter/oiofs"
	"github.com/open-io/oio-exporter/openio"
	"github.com/open-io/oio-exporter/redis"
	"github.com/open-io/oio-exporter/s3rt"
	"github.com/open-io/oio-exporter/static"
	"github.com/open-io/oio-exporter/version"
	"github.com/open-io/oio-exporter/zookeeper"
	"gopkg.in/yaml.v2"
)

var log = _log.Default
var HEALTHCHECKS = ""
var LOGPATTERNS = ""
var VERSIONS = ""

type inventory struct {
	Namespaces map[string]struct {
		Services  map[string][]*common.Service `yaml:"services"`
		Localname string                       `yaml:"localname"`
		Rack      string                       `yaml:"rack"`
	} `yaml:"namespaces"`
	Release  string `yaml:"release"`
	Hostname string `yaml:"hostname"`
}

type oioExporter struct {
	host      string
	bind      string
	inventory string
	ns        string
	localname string
	rack      string
	hcTargets []string
	metrics   *common.MetricData
	prefix    string

	state *common.State
}

// Export prometheus metrics
func (exp *oioExporter) export(w http.ResponseWriter, r *http.Request) {
	filter := r.URL.Query().Get("filter")
	exp.metrics.RLock()
	for _, metric := range exp.metrics.Measures {
		if filter != "" && !strings.HasPrefix(metric.Name, filter) {
			continue
		}

		labelStr := []string{exp.host}
		for _, labelPair := range metric.Labels {
			labelStr = append(labelStr, labelPair[0]+"=\""+labelPair[1]+"\"")
		}

		if metric.Accumulator != nil {
			for k, v := range metric.Accumulator.Values() {
				fmt.Fprintf(w, "%s%s{%s} %f\n", exp.prefix, metric.Name+"_"+k, strings.Join(labelStr, ","), v)
			}
		} else {
			fmt.Fprintf(w, "%s%s{%s} %f\n", exp.prefix, metric.Name, strings.Join(labelStr, ","), metric.Value)
		}
	}
	exp.metrics.RUnlock()
}

// Reload local service configuration
func (exp *oioExporter) reload(w http.ResponseWriter, r *http.Request) {
	n, err := exp.loadLocalConfig()
	if err != nil {
		log.Fatalln(err)
	}
	log.Noticeln("Reloaded", n, "local targets")
	fmt.Fprintf(w, "OK")
}

// Serve local service configuration
func (exp *oioExporter) config(w http.ResponseWriter, r *http.Request) {
	file, err := os.Open(exp.inventory)
	if err != nil {
		log.Fatal(err)
	}
	if _, err := io.Copy(w, file); err != nil {
		log.Warnln("could not serve config", err)
	}
}

func (exp *oioExporter) parseConfig(r io.Reader) (*inventory, error) {
	var inv *inventory
	if err := yaml.NewDecoder(r).Decode(&inv); err != nil {
		return nil, err
	}

	if inv == nil || len(inv.Namespaces) == 0 {
		return nil, errors.New("Empty inventory")
	}

	for _, servicesMap := range inv.Namespaces {
		for serviceType, services := range servicesMap.Services {
			for _, service := range services {
				service.Type = serviceType
			}
		}
	}

	return inv, nil
}

func (exp *oioExporter) loadLocalConfig() (int, error) {
	file, err := os.Open(exp.inventory)
	if err != nil {
		return 0, err
	}
	defer file.Close()

	inv, err := exp.parseConfig(file)
	if err != nil {
		return 0, err
	}

	var namespace string
	for ns := range inv.Namespaces {
		namespace = ns
		break
	}

	exp.ns = namespace
	exp.localname = inv.Namespaces[namespace].Localname
	exp.rack = inv.Namespaces[namespace].Rack
	n := len(inv.Namespaces[namespace].Services)

	exp.state.OnServicesChange.L.Lock()
	exp.state.ServicesLastUpdate = time.Now()
	exp.state.Services = inv.Namespaces[namespace].Services
	exp.state.OnServicesChange.Broadcast()
	exp.state.OnServicesChange.L.Unlock()

	return n, nil
}

func (exp *oioExporter) loadRemoteConfig(target string) (int, error) {
	services := []*common.Service{}
	res, err := http.Get("http://" + target + "/config")
	if err != nil {
		return 0, err
	}
	defer res.Body.Close()

	inv, err := exp.parseConfig(res.Body)
	if err != nil {
		return 0, err
	}

	if inv.Hostname == "" {
		inv.Hostname = strings.Split(target, ":")[0]
	}

	var remoteServices common.ServicesMap
	var namespace string
	for ns, services := range inv.Namespaces {
		remoteServices = services.Services
		namespace = ns
		break
	}
	for svcType, instances := range remoteServices {
		for _, svc := range instances {
			svc.Type = svcType
			svc.Namespace = namespace
			svc.Hostname = inv.Hostname
			services = append(services, svc)
		}
	}

	n := len(services)

	exp.state.OnTargetsChange.L.Lock()
	exp.state.TargetsLastUpdate = time.Now()
	exp.state.Targets = append(exp.state.Targets, services...)
	exp.state.OnTargetsChange.Broadcast()
	exp.state.OnTargetsChange.L.Unlock()

	return n, nil
}

func (exp *oioExporter) reloadRemote() {
	retry := 10 * time.Second

	var wg sync.WaitGroup
	wg.Add(len(exp.hcTargets))

	for _, target := range exp.hcTargets {
		go func(target string) {
			defer wg.Done()
			for {
				n, err := exp.loadRemoteConfig(target)
				if err != nil {
					log.Warnf("Error loading remote config '%s': %s; retrying in %s", target, err, retry)
					time.Sleep(retry)
					continue
				}

				log.Noticeln("Loaded", n, "remote targets")

				return
			}
		}(target)
	}

	wg.Wait()

	log.Notice("Loaded all remote targets")
}

func (exp *oioExporter) listen(stop <-chan struct{}) {
	http.HandleFunc("/metrics", exp.export)
	http.HandleFunc("/config", exp.config)
	http.HandleFunc("/reload", exp.reload)
	server := &http.Server{Addr: exp.bind}

	go func() {
		if err := server.ListenAndServe(); err != nil && !errors.Is(err, http.ErrServerClosed) {
			log.Fatalf("HTTP server error: %s", err)
		}
	}()

	<-stop

	if err := server.Shutdown(context.TODO()); err != nil {
		log.Notice(err)
	}
}

func main() {
	var inventory string
	var targets string
	var patternPath string
	var healthcheckPath string
	var versionPath string
	var bind string
	var daemonsToRun string
	var daemonsNotToRun string

	host, err := os.Hostname()
	if err != nil {
		log.Warnln("Could not retrieve hostname", err)
	}

	fs := flag.NewFlagSet("", flag.ExitOnError)
	fs.StringVar(&inventory, "inventory", "/etc/oio/sds/OPENIO/inventory.yml", "Path to inventory file")
	fs.StringVar(&patternPath, "pattern-path", "", "Path to log patterns")
	fs.StringVar(&healthcheckPath, "healthcheck-path", "", "Path to healthchecks definition")
	fs.StringVar(&versionPath, "version-path", "", "Path to versions definition")
	fs.StringVar(&host, "host", host, "Report metrics with the specified hostname")
	fs.StringVar(&bind, "bind", "0.0.0.0:6920", "Listen address")
	fs.StringVar(&targets, "targets", "localhost:6920", "Comma separated list of targets to perform healthchecks on")
	fs.StringVar(&daemonsToRun, "daemons", "", "Comma separated list of daemons to run (empty means all of them)")
	fs.StringVar(&daemonsNotToRun, "nodaemons", "", "Comma separated list of daemons to explicitly disable (empty means nothing to disable)")

	_log.InitFlags(fs)

	if err := fs.Parse(os.Args[1:]); err != nil {
		log.Fatalln("Could not parse args", err)
	}

	if err := _log.Setup(); err != nil {
		log.Fatal(err)
	}

	exp := oioExporter{
		host:      fmt.Sprintf(`host="%s"`, host),
		bind:      bind,
		inventory: inventory,
		hcTargets: strings.Split(targets, ","),
		metrics:   common.MakeMetricData(),
		prefix:    "openio_",
		state:     common.NewState(),
	}

	n, err := exp.loadLocalConfig()
	if err != nil {
		log.Fatalln("in loading local config", err)
	}
	log.Noticeln("Loaded", n, "local targets")

	go exp.reloadRemote()

	allDaemonNames := "hc,oio,redis,version,s3rt,log,oiofs,self,static,memcached,beanstalkd,zookeeper"
	if len(daemonsToRun) == 0 {
		daemonsToRun = allDaemonNames
	}

	daemonsToRunNames := strings.Split(daemonsToRun, ",")
	daemonsNotToRunNames := strings.Split(daemonsNotToRun, ",")

	daemons := []daemon.DaemonI{}
	for _, daemonName := range daemonsToRunNames {
		if common.Index(daemonsNotToRunNames, daemonName) > -1 {
			log.Debugln("Ignoring daemon", daemonName, "...")
			continue
		}
		switch daemonName {
		case "hc":
			daemons = append(daemons, hc.MakeDaemon(HEALTHCHECKS, healthcheckPath))
		case "oio":
			daemons = append(daemons, openio.MakeDaemon(exp.ns))
		case "redis":
			daemons = append(daemons, redis.MakeDaemon(exp.ns))
		case "version":
			daemons = append(daemons, version.MakeDaemon(VERSIONS, versionPath, exp.ns))
		case "s3rt":
			daemons = append(daemons, s3rt.MakeDaemon(exp.ns))
		case "log":
			daemons = append(daemons, clog.MakeDaemon(LOGPATTERNS, patternPath, 10*time.Second, exp.ns))
		case "oiofs":
			daemons = append(daemons, oiofs.MakeDaemon(exp.ns))
		case "self":
			daemons = append(daemons, self.MakeDaemon())
		case "static":
			daemons = append(daemons, static.MakeDaemon(exp.ns, exp.localname, exp.rack))
		case "memcached":
			daemons = append(daemons, memcached.MakeDaemon(exp.ns))
		case "beanstalkd":
			daemons = append(daemons, beanstalkd.MakeDaemon(exp.ns))
		case "zookeeper":
			daemons = append(daemons, zookeeper.MakeDaemon(exp.ns))
		default:
			log.Fatalf("ERROR Unknown daemon %s (daemons are: %s)", daemonName, allDaemonNames)
		}
	}
	if len(daemons) == 0 {
		log.Fatalf("ERROR no daemon selected, at least one is necessary to run.")
	}

	sigCh := make(chan os.Signal, 1)
	signal.Notify(sigCh, os.Interrupt, syscall.SIGTERM, syscall.SIGHUP)

	stop := make(chan struct{})
	var wg sync.WaitGroup
	wg.Add(1 + len(daemons))

	go func() {
		defer wg.Done()
		exp.listen(stop)
	}()

	for _, d := range daemons {
		go func(d daemon.DaemonI) {
			defer wg.Done()

			log.Noticef("start %s", d.ID())
			if err := daemon.Run(d, exp.metrics, exp.state, stop); err != nil {
				// TODO: match Init errors with errors.Is & retry, return only for Close errors
				log.Errorf("%s: %s", d.ID(), err)

				return
			}
			log.Infof("Closed %s", d.ID())
		}(d)
	}

	<-sigCh

	log.Notice("Closing gracefully")

	close(stop)

	// Free the reloading goroutines
	exp.state.OnServicesChange.L.Lock()
	exp.state.OnServicesChange.Broadcast()
	exp.state.OnServicesChange.L.Unlock()
	exp.state.OnTargetsChange.L.Lock()
	exp.state.OnTargetsChange.Broadcast()
	exp.state.OnTargetsChange.L.Unlock()

	waitCh := make(chan struct{})
	go func() {
		defer close(waitCh)
		wg.Wait()
	}()

	select {
	case <-waitCh:
		log.Notice("Exited")
	case <-time.After(20 * time.Second):
		log.Fatal("Timeout")
	case <-sigCh:
		log.Fatal("Forced Exit")
	}
}
